// Config: parse hosts.yaml

// Copyright (C) 2021  Simon Ruderich
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package config

import (
	"fmt"
	"os"
	"strings"

	"gopkg.in/yaml.v2"
)

type Hosts struct {
	List []*Host
	Map  map[string]*Host
}

type Host struct {
	Name string `yaml:"name"`

	SshUser string `yaml:"ssh_user"`
}

func LoadHosts() (*Hosts, error) {
	const path = "hosts.yaml"

	var hostList []*Host
	x, err := os.ReadFile(path)
	if err != nil {
		return nil, err
	}
	err = yaml.UnmarshalStrict(x, &hostList)
	if err != nil {
		return nil, fmt.Errorf("%s: failed to load: %v", path, err)
	}

	hostMap := make(map[string]*Host)
	for _, x := range hostList {
		errPrefix := fmt.Sprintf("%s: host %q:", path, x.Name)
		if x.Name == GroupAll {
			return nil, fmt.Errorf(
				"%s conflict with pre-defined group %q",
				errPrefix, x.Name)
		}
		if strings.HasPrefix(x.Name, GroupDetectedPrefix) {
			return nil, fmt.Errorf(
				"%s name must not start with %q "+
					"(reserved for detected groups)",
				errPrefix, GroupDetectedPrefix)
		}
		if strings.Contains(x.Name, GroupSpecialSeparator) {
			return nil, fmt.Errorf(
				"%s name must not contain %q",
				errPrefix, GroupSpecialSeparator)
		}

		if hostMap[x.Name] != nil {
			return nil, fmt.Errorf("%s host name already exists",
				errPrefix)
		}
		hostMap[x.Name] = x
	}

	return &Hosts{
		List: hostList,
		Map:  hostMap,
	}, nil
}
